# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=too-many-lines

import logging
from datetime import date
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Set, cast

import ezodf

from rp2.abstract_country import AbstractCountry
from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.balance import BalanceSet
from rp2.computed_data import ComputedData, YearlyGainLoss
from rp2.gain_loss import GainLoss
from rp2.gain_loss_set import GainLossSet
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.localization import _
from rp2.logger import create_logger
from rp2.out_transaction import OutTransaction
from rp2.plugin.report.abstract_ods_generator import AbstractODSGenerator
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError
from rp2.transaction_set import TransactionSet

LOGGER: logging.Logger = create_logger("rp2_full_report")


class _TransactionVisualStyle(NamedTuple):
    year: int
    visual_style: str
    highlighted_style: str


class _BorderStyle(NamedTuple):
    year: int
    border_suffix: str


class _AssetAndYear(NamedTuple):
    asset: str
    year: int


_ZERO: RP2Decimal = RP2Decimal(0)


class Generator(AbstractODSGenerator):
    MIN_ROWS: int = 40
    MAX_COLUMNS: int = 40
    OUTPUT_FILE: str = "rp2_full_report.ods"

    TEMPLATE_SHEETS_TO_KEEP: Set[str] = {"__Summary"}

    __in_out_sheet_transaction_2_row: Dict[AbstractTransaction, int] = {}
    __tax_sheet_year_2_row: Dict[_AssetAndYear, int] = {}

    __legend: List[List[str]] = []
    __yearly_gain_loss_summary_header_names_row_1: List[str] = []
    __yearly_gain_loss_summary_header_names_row_2: List[str] = []
    __in_header_names_row_1: List[str] = []
    __in_header_names_row_2: List[str] = []
    __out_header_names_row_1: List[str] = []
    __out_header_names_row_2: List[str] = []
    __intra_header_names_row_1: List[str] = []
    __intra_header_names_row_2: List[str] = []
    __balance_header_names_row_1: List[str] = []
    __balance_header_names_row_2: List[str] = []
    __gain_loss_summary_header_names_row_1: List[str] = []
    __gain_loss_summary_header_names_row_2: List[str] = []
    __gain_loss_detail_header_names_row_1: List[str] = []
    __gain_loss_detail_header_names_row_2: List[str] = []

    # pylint: disable=line-too-long
    def _setup_text_data(self, country: AbstractCountry) -> None:
        currency_code: str = country.currency_iso_code.upper()

        self.__legend: List[List[str]] = [
            # fmt: off
            [_("General")],
            [_("Accounting Method")],
            [_("From Date Filter")],
            [_("To Date Filter")],
            [""],
            [_("Sheet Types")],
            [_("<Crypto> In-Out"), _("Captures all transactions coming in (IN), going out (OUT) and transferring across accounts (INTRA)")],
            [_("<Crypto> Tax"), _("Computation of balances and gain / loss")],
            [""],
            [_("Table Types")],
            [_("In-Flow Detail"), _("Transactions that added new crypto to accounts (e.g. buy, etc.). Only EARN-typed transactions are taxable events (interest received on crypto)")],
            [_("Out-Flow Detail"), _("Transactions that removed some crypto from accounts (e.g. sell, send as gift, etc.). These are taxable events")],
            [_("Intra-Flow Detail"), _("Movements across accounts without increasing/decreasing total value of crypto owned. The amount transferred is non-taxable but the transfer fee is considered a crypto sale and it is a taxable event")],
            [_("Gain / Loss Summary"), _("Computed gain and loss for the given cryptocurrency, organized by year and capital gains type (LONG or SHORT)")],
            [_("Account Balances"), _("Computed balances of all accounts. Useful to double-check that the input transactions have been entered correctly. If values don’t match actual balances some data is missing or wrong")],
            [_("Average Price"), _("Average price at which the crypto was acquired")],
            [_("Gain / Loss Detail"), _("Detailed computation of gain and loss: each lot is divided into fractions, which are used to calculate the cost basis and the gain/loss")],
            [""],
            [_("In-Flow Detail")],
            [_("Sent/Sold"), _("Lots that have been sent or sold, according to the order defined by the Accounting Method (see General section)")],
            [_("Timestamp"), _("Time at which the transaction occurred")],
            [_("Asset"), _("Which cryptocurrency was transacted (e.g. BTC, ETH, etc.)")],
            [_("Exchange"), _("Exchange or wallet on which the transaction occurred (e.g. Coinbase, Coinbase Pro, BlockFi, etc.)")],
            [_("Holder"), _("Exchange account or wallet owner")],
            [_("Transaction Type"), _("Type of the transaction (BUY, GIFT, INTEREST, STAKING, etc.)")],
            [_("Spot Price"), _("Value of 1 unit of the given cryptocurrency at the time the transaction occurred")],
            [_("Crypto In"), _("How much of the given cryptocurrency was acquired with the transaction")],
            [_("Crypto In Running Sum"), _("Running sum of crypto received")],
            [_("{} Fee").format(currency_code), _("{} value of the fees").format(currency_code)],
            [_("{} In No Fee").format(currency_code), _("{} value of the transaction without fees").format(currency_code)],
            [_("{} In With Fee").format(currency_code), _("{} value of the transaction with fees").format(currency_code)],
            [_("Taxable Event"), _("Does the transaction contain a taxable event? If so the taxable amount is highlighted in yellow")],
            [_("Unique Id"), _("Hash or exchange-specific unique identifier for the transaction")],
            [_("Notes"), _("Description of the transaction")],
            [""],
            [_("Out-Flow Detail")],
            [_("Timestamp"), _("Time at which the transaction occurred")],
            [_("Asset"), _("Which cryptocurrency was transacted (e.g. BTC, ETH, etc.)")],
            [_("Exchange"), _("Exchange or wallet on which the transaction occurred (e.g. Coinbase, Coinbase Pro, BlockFi, etc.)")],
            [_("Holder"), _("Exchange account or wallet owner")],
            [_("Transaction Type"), _("Type of the transaction (DONATE, GIFT, SELL, etc.)")],
            [_("Spot Price"), _("Value of 1 unit of the given cryptocurrency at the time the transaction occurred")],
            [_("Crypto Out"), _("How much of the given cryptocurrency was sent with the transaction")],
            [_("Crypto Fee"), _("Crypto value of the fees")],
            [_("Crypto Out Running Sum"), _("Running sum of crypto sent")],
            [_("Crypto Fee Running Sum"), _("Running sum of crypto fees")],
            [_("{} Out").format(currency_code), _("{} value of the transaction without fees").format(currency_code)],
            [_("{} Fee").format(currency_code), _("{} value of the fees").format(currency_code)],
            [_("Taxable Event"), _("Does the transaction contain a taxable event? If so the taxable amount is highlighted in yellow")],
            [_("Unique Id"), _("Hash or exchange-specific unique identifier for the transaction")],
            [_("Notes"), _("Description of the transaction")],
            [""],
            [_("Intra-Flow Detail")],
            [_("Timestamp"), _("Time at which the transaction occurred")],
            [_("Asset"), _("Which cryptocurrency was transacted (e.g. BTC, ETH, etc.)")],
            [_("From Exchange"), _("Exchange or wallet from which the transfer of crypto occurred (e.g. Coinbase, Coinbase Pro, BlockFi, etc.)")],
            [_("From Holder"), _("Owner of the exchange account or wallet from which the transfer of crypto occurred")],
            [_("To Exchange"), _("Exchange or wallet to which the transfer of crypto occurred (e.g. Coinbase, Coinbase Pro, BlockFi, etc.)")],
            [_("To Holder"), _("Owner of the exchange account or wallet to which the transfer of crypto occurred")],
            [_("Spot Price"), _("Value of 1 unit of the given cryptocurrency at the time the transaction occurred")],
            [_("Crypto Sent"), _("How much of the given cryptocurrency was sent with the transaction")],
            [_("Crypto Received"), _("How much of the given cryptocurrency was received with the transaction")],
            [_("Crypto Fee"), _("Crypto value of the fees")],
            [_("Crypto Fee Running Sum"), _("Running sum of crypto fees")],
            [_("{} Fee").format(currency_code), _("{} value of the fees").format(currency_code)],
            [_("Taxable Event"), _("Does the transaction contain a taxable event? If so the taxable amount is highlighted in yellow")],
            [_("Unique Id"), _("Hash of the transaction")],
            [_("Notes"), _("Description of the transaction")],
            [""],
            [_("Gain / Loss Summary")],
            [_("Year"), _("Summary year")],
            [_("Asset"), _("Which cryptocurrency (e.g. BTC, ETH, etc.)")],
            [_("Capital Gains"), _("Sum of all capital gains in {} for transactions of the given capital gains type").format(currency_code)],
            [_("Capital Gains Type"), _("LONG (> 1 year) or SHORT (< 1 year)")],
            [_("Transaction Type"), _("EARN (crypto earned through interest, etc.), GIFT (crypto given), SOLD (crypto sold) OR INTRA (fees for transferring crypto across accounts)")],
            [_("Crypto Taxable Total"), _("Sum of all taxable events in crypto for transactions of the given capital gains type")],
            [_("{} Taxable Total").format(currency_code), _("Sum of all taxable events in {} for transactions of the given capital gains type").format(currency_code)],
            [_("{} Total Cost Basis").format(currency_code), _("Sum of all cost bases in {} for transactions of the given capital gains type").format(currency_code)],
            [""],
            [_("Account Balances")],
            [_("Exchange"), _("Exchange or wallet on which the transaction occurred (e.g. Coinbase, Coinbase Pro, BlockFi, etc.)")],
            [_("Holder"), _("Name of the exchange or wallet account holder")],
            [_("Asset"), _("Which cryptocurrency was transacted (e.g. BTC, ETH, etc.)")],
            [_("Acquired Balance"), _("Balance of all the BUY and EARN transactions for a given exchange and holder")],
            [_("Sent Balance"), _("Balance of all the SEND and SOLD transactions for which the given exchange and holder are sender")],
            [_("Received Balance"), _("Balance of all the SEND and SOLD transactions for which the given exchange and holder are receiver")],
            [_("Final Balance"), _("Final balance of all transactions for a given exchange and holder")],
            [""],
            [_("Average Price")],
            [_("Average Price Paid Per 1 crypto"), _("Average price at which the crypto was acquired, across all transactions")],
            [""],
            [_("Gain / Loss Detail")],
            [_("Crypto Amount"), _("Crypto amount for the given taxable event fraction")],
            [_("Asset"), _("Which cryptocurrency (e.g. BTC, ETH, etc.)")],
            [_("Crypto Amt Running Sum"), _("Running sum of crypto amount")],
            [_("Capital Gains"), _("Sum of all capital gains in {} for transactions of the given capital gains type").format(currency_code)],
            [_("Capital Gains Type"), _("LONG (> 1 year) or SHORT (< 1 year)")],
            [_("Taxable Event Timestamp"), _("Time at which the taxable event occurred")],
            [_("Taxable Event Direction/Type"), _("Direction (IN/OUT/INTRA) and type (BUY, SOLD, etc) of the taxable event")],
            [_("Taxable Event Fraction %"), _("Percentage of the taxable event")],
            [_("Taxable Event {} Amount Fraction").format(currency_code), _("{} amount of this taxable event fraction").format(currency_code)],
            [_("Taxable Event Spot Price"), _("Value of 1 unit of the given cryptocurrency at the time the taxable event occurred")],
            [_("Taxable Event Fraction Description"), _("English description of this taxable event fraction")],
            [_("Acquired Lot Timestamp"), _("Time at which the in-transaction occurred")],
            [_("Acquired Lot Fraction %"), _("Percentage of the in-transaction")],
            [_("Acquired Lot {} Amount Fraction").format(currency_code), _("{} amount of this in-transaction fraction").format(currency_code)],
            [_("Acquired Lot {} Fee Fraction").format(currency_code), _("{} fee of this in-transaction fraction").format(currency_code)],
            [_("Acquired Lot {} Cost Basis").format(currency_code), _("{} cost basis of this in-transaction fraction").format(currency_code)],
            [_("Acquired Lot Spot Price"), _("Value of 1 unit of the given cryptocurrency at the time the in-transaction occurred")],
            [_("Acquired Lot Fraction Description"), _("English description of this in-transaction fraction")],
            # fmt: on
        ]

        self.__yearly_gain_loss_summary_header_names_row_1: List[str] = [
            "",
            "",
            _("Capital"),
            _("Capital"),
            _("Transaction"),
            _("Crypto"),
            _("USD"),
            _("USD Total"),
        ]

        self.__yearly_gain_loss_summary_header_names_row_2: List[str] = [
            _("Year"),
            _("Asset"),
            _("Gains"),
            _("Gains Type"),
            _("Type"),
            _("Taxable Total"),
            _("Taxable Total"),
            _("Cost Basis"),
        ]

        self.__in_header_names_row_1: List[str] = [
            "",
            "",
            "",
            "",
            "",
            _("Transaction"),
            "",
            _("Crypto"),
            _("Crypto In"),
            "",
            _("{} In").format(currency_code),
            _("{} In").format(currency_code),
            _("Taxable"),
            "",
            "",
            "",
        ]

        self.__in_header_names_row_2: List[str] = [
            _("Sent/Sold"),
            _("Timestamp"),
            _("Asset"),
            _("Exchange"),
            _("Holder"),
            _("Type"),
            _("Spot Price"),
            _("In"),
            _("Running Sum"),
            _("{} Fee").format(currency_code),
            _("No Fee"),
            _("With Fee"),
            _("Event"),
            _("N/A"),
            _("Unique Id"),
            _("Notes"),
        ]

        self.__out_header_names_row_1: List[str] = [
            "",
            "",
            "",
            "",
            _("Transaction"),
            "",
            "",
            "",
            _("Crypto Out"),
            _("Crypto Fee"),
            "",
            "",
            _("Taxable"),
            "",
            "",
        ]

        self.__out_header_names_row_2: List[str] = [
            _("Timestamp"),
            _("Asset"),
            _("Exchange"),
            _("Holder"),
            _("Type"),
            _("Spot Price"),
            _("Crypto Out"),
            _("Crypto Fee"),
            _("Running Sum"),
            _("Running Sum"),
            _("{} Out").format(currency_code),
            _("{} Fee").format(currency_code),
            _("Event"),
            _("Unique Id"),
            _("Notes"),
        ]

        self.__intra_header_names_row_1: List[str] = [
            "",
            "",
            _("From"),
            _("From"),
            "",
            "",
            "",
            "",
            _("Crypto"),
            "",
            _("Crypto Fee"),
            "",
            _("Taxable"),
            "",
            "",
        ]

        self.__intra_header_names_row_2: List[str] = [
            _("Timestamp"),
            _("Asset"),
            _("Exchange"),
            _("Holder"),
            _("To Exchange"),
            _("To Holder"),
            _("Spot Price"),
            _("Crypto Sent"),
            _("Received"),
            _("Crypto Fee"),
            _("Running Sum"),
            _("{} Fee").format(currency_code),
            _("Event"),
            _("Unique Id"),
            _("Notes"),
        ]

        self.__balance_header_names_row_1: List[str] = [
            "",
            "",
            "",
            _("Acquired"),
            _("Sent"),
            _("Received"),
            _("Final"),
        ]

        self.__balance_header_names_row_2: List[str] = [
            _("Exchange"),
            _("Holder"),
            _("Asset"),
            _("Balance"),
            _("Balance"),
            _("Balance"),
            _("Balance"),
        ]

        self.__gain_loss_summary_header_names_row_1: List[str] = [
            "",
            "",
            _("Capital"),
            _("Capital"),
            _("Transaction"),
            _("Crypto"),
            _("{}").format(currency_code),
            _("{} Total").format(currency_code),
        ]

        self.__gain_loss_summary_header_names_row_2: List[str] = [
            _("Year"),
            _("Asset"),
            _("Gains"),
            _("Gains Type"),
            _("Type"),
            _("Taxable Total"),
            _("Taxable Total"),
            _("Cost Basis"),
        ]

        self.__gain_loss_detail_header_names_row_1: List[str] = [
            _("Crypto"),
            "",
            _("Crypto Amt"),
            _("Capital"),
            _("Capital"),
            _("Taxable Event"),
            _("Taxable Event"),
            _("Taxable Event"),
            _("Taxable Event {}").format(currency_code),
            _("Taxable Event"),
            "",
            _("Taxable Event"),
            _("Acquired Lot"),
            _("Acquired Lot"),
            _("Acquired Lot {}").format(currency_code),
            _("Acquired Lot {}").format(currency_code),
            _("Acquired Lot {}").format(currency_code),
            _("Acquired Lot"),
            "",
            _("Acquired Lot Fraction"),
        ]

        self.__gain_loss_detail_header_names_row_2: List[str] = [
            _("Amount"),
            _("Asset"),
            _("Running Sum"),
            _("Gains"),
            _("Gains Type"),
            _("Timestamp"),
            _("Direction/Type"),
            _("Fraction %"),
            _("Amount Fraction"),
            _("Spot Price"),
            _("Unique Id"),
            _("Fraction Description"),
            _("Timestamp"),
            _("Fraction %"),
            _("Amount Fraction"),
            _("Fee Fraction"),
            _("Cost Basis"),
            _("Spot Price"),
            _("Unique Id"),
            _("Description"),
        ]

    def generate(
        self,
        country: AbstractCountry,
        years_2_accounting_method_names: Dict[int, str],
        asset_to_computed_data: Dict[str, ComputedData],
        output_dir_path: str,
        output_file_prefix: str,
        from_date: date,
        to_date: date,
        generation_language: str,
    ) -> None:
        if not isinstance(asset_to_computed_data, Dict):
            raise RP2TypeError(f"Parameter 'asset_to_computed_data' has non-Dict value {asset_to_computed_data}")

        self._setup_text_data(country)

        template_path: str = self._get_template_path("rp2_full_report", country, generation_language)

        output_file: Any
        output_file = self._initialize_output_file(
            country=country,
            legend_data=self.__legend,
            years_2_accounting_method_names=years_2_accounting_method_names,
            output_dir_path=output_dir_path,
            output_file_prefix=output_file_prefix,
            output_file_name=self.OUTPUT_FILE,
            template_path=template_path,
            template_sheets_to_keep=self.TEMPLATE_SHEETS_TO_KEEP,
            from_date=from_date,
            to_date=to_date,
        )
        asset: str
        computed_data: ComputedData

        summary_sheet = output_file.sheets["Summary"]
        summary_row_index: int = self._fill_header(
            _("Yearly Gain / Loss Summary"),
            self.__yearly_gain_loss_summary_header_names_row_1,
            self.__yearly_gain_loss_summary_header_names_row_2,
            summary_sheet,
            0,
            0,
        )
        for asset, computed_data in asset_to_computed_data.items():
            if not isinstance(asset, str):
                raise RP2TypeError(f"Parameter 'asset' has non-string value {asset}")
            ComputedData.type_check("computed_data", computed_data)
            summary_row_index = self.__generate_asset(computed_data, output_file, summary_row_index)

        summary_sheet.name = _("Summary")

        output_file.save()
        LOGGER.info("Plugin '%s' output: %s", __name__, Path(output_file.docname).resolve())

    @staticmethod
    def get_in_out_sheet_name(asset: str) -> str:
        return _("{} In-Out").format(asset)

    @staticmethod
    def get_tax_sheet_name(asset: str) -> str:
        return _("{} Tax").format(asset)

    def __get_number_of_rows_in_transaction_sheet(self, computed_data: ComputedData) -> int:
        return self.MIN_ROWS + computed_data.in_transaction_set.count + computed_data.out_transaction_set.count + computed_data.intra_transaction_set.count

    def __get_number_of_rows_in_output_sheet(self, computed_data: ComputedData) -> int:
        return self.MIN_ROWS + len(computed_data.yearly_gain_loss_list) + computed_data.balance_set.count + computed_data.gain_loss_set.count

    def __generate_asset(self, computed_data: ComputedData, output_file: Any, summary_row_index: int) -> int:
        asset: str = computed_data.asset
        transaction_sheet_name: str = self.get_in_out_sheet_name(asset)
        output_sheet_name: str = self.get_tax_sheet_name(asset)

        transaction_sheet: Any = ezodf.Table(transaction_sheet_name)
        output_sheet: Any = ezodf.Table(output_sheet_name)
        summary_sheet: Any = output_file.sheets["Summary"]

        output_file.sheets += transaction_sheet
        output_file.sheets += output_sheet

        transaction_sheet.reset(size=(self.__get_number_of_rows_in_transaction_sheet(computed_data), self.MAX_COLUMNS))
        output_sheet.reset(size=(self.__get_number_of_rows_in_output_sheet(computed_data), self.MAX_COLUMNS))

        new_lines: int = len(computed_data.yearly_gain_loss_list)
        if new_lines:
            summary_sheet.append_rows(new_lines)

        row_index: int = 0
        row_index = self.__generate_in_table(transaction_sheet, computed_data, row_index)
        row_index = self.__generate_out_table(transaction_sheet, computed_data, row_index + 2)
        row_index = self.__generate_intra_table(transaction_sheet, computed_data, row_index + 2)

        row_index = 0
        row_index = self.__generate_gain_loss_summary(output_sheet, computed_data.yearly_gain_loss_list, row_index)
        row_index = self.__generate_account_balances(output_sheet, computed_data.balance_set, row_index + 2)
        row_index = self.__generate_average_price_per_unit(output_sheet, asset, computed_data.price_per_unit, row_index + 2)
        row_index = self.__generate_gain_loss_detail(output_sheet, asset, computed_data, row_index + 2)

        return self.__generate_yearly_gain_loss_summary(summary_sheet, asset, computed_data.yearly_gain_loss_list, summary_row_index)

    @staticmethod
    def __get_transaction_visual_style(transaction: AbstractTransaction, year: int) -> _TransactionVisualStyle:
        visual_style: str = "transparent"
        highlighted_style: str = "transparent"
        if transaction.is_taxable():
            visual_style = "taxable_event"
            highlighted_style = "highlighted"
        if year == 0:
            year = transaction.timestamp.year
        if transaction.timestamp.year != year:
            visual_style = f"{visual_style}_border"
            highlighted_style = f"{highlighted_style}_border"
            year = transaction.timestamp.year
        return _TransactionVisualStyle(year, visual_style, highlighted_style)

    @staticmethod
    def __get_border_style(current_year: int, year: int) -> _BorderStyle:
        border_suffix: str = ""
        if year == 0:
            year = current_year
        if current_year != year:
            border_suffix = "_border"
            year = current_year
        return _BorderStyle(year, border_suffix)

    def __generate_in_table(self, sheet: Any, computed_data: ComputedData, row_index: int) -> int:
        row_index = self._fill_header(_("In-Flow Detail"), self.__in_header_names_row_1, self.__in_header_names_row_2, sheet, row_index, 0)

        in_transaction_set: TransactionSet = computed_data.in_transaction_set
        entry: AbstractEntry
        year: int = 0
        visual_style: str
        previous_transaction: Optional[InTransaction] = None
        border_style: _BorderStyle
        border_suffix: str = ""
        for entry in in_transaction_set:
            transaction: InTransaction = cast(InTransaction, entry)
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            border_style = self.__get_border_style(entry.timestamp.year, year)
            border_suffix = border_style.border_suffix
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            in_lot_sold_percentage: Optional[RP2Decimal] = computed_data.get_in_lot_sold_percentage(transaction)

            # Write _ZERO only on the first in_transaction if there are no sold lots
            if in_lot_sold_percentage == _ZERO and previous_transaction is not None:
                in_lot_sold_percentage = None
            self._fill_cell(
                sheet,
                row_index,
                0,
                in_lot_sold_percentage if in_lot_sold_percentage is not None else "",
                data_style="percent",
                visual_style="acquired_lot" + border_suffix if in_lot_sold_percentage is not None else "transparent",
            )
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.transaction_type.get_translation().upper(), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.spot_price, data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 7, transaction.crypto_in, data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 8, computed_data.get_crypto_in_running_sum(transaction), data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 9, transaction.fiat_fee, data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 10, transaction.fiat_in_no_fee, data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 11, transaction.fiat_in_with_fee, data_style="fiat", visual_style=highlighted_style)
            self._fill_cell(sheet, row_index, 12, _("YES") if transaction.is_taxable() else _("NO"), data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 13, "", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.unique_id, visual_style="transparent")
            self._fill_cell(sheet, row_index, 15, transaction.notes, visual_style="transparent")

            self.__in_out_sheet_transaction_2_row[transaction] = row_index + 1

            previous_transaction = transaction
            row_index += 1

        return row_index

    def __generate_out_table(self, sheet: Any, computed_data: ComputedData, row_index: int) -> int:
        row_index = self._fill_header(_("Out-Flow Detail"), self.__out_header_names_row_1, self.__out_header_names_row_2, sheet, row_index, 1)

        out_transaction_set: TransactionSet = computed_data.out_transaction_set

        entry: AbstractEntry
        year: int = 0
        for entry in out_transaction_set:
            transaction: OutTransaction = cast(OutTransaction, entry)
            visual_style: str
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            self._fill_cell(sheet, row_index, 0, "", visual_style="transparent")
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.transaction_type.get_translation().upper(), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.spot_price, visual_style=visual_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 7, transaction.crypto_out_no_fee, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 8, transaction.crypto_fee, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 9, computed_data.get_crypto_out_running_sum(transaction), data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 10, computed_data.get_crypto_out_fee_running_sum(transaction), data_style="crypto", visual_style=visual_style)
            self._fill_cell(
                sheet,
                row_index,
                11,
                transaction.fiat_out_no_fee,
                visual_style=highlighted_style if transaction.fiat_out_no_fee > ZERO else visual_style,
                data_style="fiat",
            )
            self._fill_cell(
                sheet,
                row_index,
                12,
                transaction.fiat_fee,
                visual_style=highlighted_style if transaction.fiat_fee > ZERO else visual_style,
                data_style="fiat",
            )
            self._fill_cell(sheet, row_index, 13, _("YES") if transaction.is_taxable() else _("NO"), data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.unique_id, visual_style="transparent")
            self._fill_cell(sheet, row_index, 15, transaction.notes, visual_style="transparent")

            self.__in_out_sheet_transaction_2_row[transaction] = row_index + 1

            row_index += 1

        return row_index

    def __generate_intra_table(self, sheet: Any, computed_data: ComputedData, row_index: int) -> int:
        row_index = self._fill_header(_("Intra-Flow Detail"), self.__intra_header_names_row_1, self.__intra_header_names_row_2, sheet, row_index, 1)

        intra_transaction_set: TransactionSet = computed_data.intra_transaction_set

        entry: AbstractEntry
        year: int = 0
        for entry in intra_transaction_set:
            transaction: IntraTransaction = cast(IntraTransaction, entry)
            visual_style: str
            highlighted_style: str
            transaction_visual_style: _TransactionVisualStyle = self.__get_transaction_visual_style(transaction, year)
            year = transaction_visual_style.year
            visual_style = transaction_visual_style.visual_style
            highlighted_style = transaction_visual_style.highlighted_style
            self._fill_cell(sheet, row_index, 0, "", visual_style="transparent")
            self._fill_cell(sheet, row_index, 1, transaction.timestamp, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 2, transaction.asset, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 3, transaction.from_exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 4, transaction.from_holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 5, transaction.to_exchange, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 6, transaction.to_holder, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 7, transaction.spot_price, visual_style=visual_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 8, transaction.crypto_sent, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 9, transaction.crypto_received, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 10, transaction.crypto_fee, visual_style=visual_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 11, computed_data.get_crypto_intra_fee_running_sum(transaction), data_style="crypto", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 12, transaction.fiat_fee, visual_style=highlighted_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 13, _("YES") if transaction.is_taxable() else _("NO"), data_style="fiat", visual_style=visual_style)
            self._fill_cell(sheet, row_index, 14, transaction.unique_id, visual_style=visual_style)
            self._fill_cell(sheet, row_index, 15, transaction.notes, visual_style="transparent")

            self.__in_out_sheet_transaction_2_row[transaction] = row_index + 1

            row_index += 1

        return row_index

    def __generate_gain_loss_summary(self, sheet: Any, yearly_gain_loss_list: List[YearlyGainLoss], row_index: int) -> int:
        row_index = self._fill_header(
            _("Gain / Loss Summary"), self.__gain_loss_summary_header_names_row_1, self.__gain_loss_summary_header_names_row_2, sheet, row_index, 0
        )

        year: int = 0
        for yearly_gain_loss in yearly_gain_loss_list:
            border_suffix: str = ""
            capital_gains_type: str = _("LONG") if yearly_gain_loss.is_long_term_capital_gains else _("SHORT")
            border_style: _BorderStyle = self.__get_border_style(yearly_gain_loss.year, year)
            year = border_style.year
            border_suffix = border_style.border_suffix
            self._fill_cell(sheet, row_index, 0, yearly_gain_loss.year, visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 1, yearly_gain_loss.asset, visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 2, yearly_gain_loss.fiat_gain_loss, visual_style="bold" + border_suffix, data_style="fiat")
            self._fill_cell(sheet, row_index, 3, capital_gains_type, visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(
                sheet, row_index, 4, yearly_gain_loss.transaction_type.get_translation().upper(), visual_style="bold" + border_suffix, data_style="default"
            )
            self._fill_cell(sheet, row_index, 5, yearly_gain_loss.crypto_amount, visual_style="transparent" + border_suffix, data_style="crypto")
            self._fill_cell(
                sheet,
                row_index,
                6,
                yearly_gain_loss.fiat_amount,
                visual_style="taxable_event" + border_suffix,
                data_style="fiat",
            )
            self._fill_cell(
                sheet,
                row_index,
                7,
                yearly_gain_loss.fiat_cost_basis,
                visual_style="acquired_lot" + border_suffix,
                data_style="fiat",
            )
            row_index += 1

        return row_index

    def __generate_account_balances(self, sheet: Any, balance_set: BalanceSet, row_index: int) -> int:
        row_index = self._fill_header(_("Account Balances"), self.__balance_header_names_row_1, self.__balance_header_names_row_2, sheet, row_index, 0)

        totals: Dict[str, RP2Decimal] = {}
        value: RP2Decimal
        for balance in balance_set:
            self._fill_cell(sheet, row_index, 0, balance.exchange, visual_style="bold", data_style="default")
            self._fill_cell(sheet, row_index, 1, balance.holder, visual_style="bold", data_style="default")
            self._fill_cell(sheet, row_index, 2, balance.asset)
            self._fill_cell(sheet, row_index, 3, balance.acquired_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 4, balance.sent_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 5, balance.received_balance, data_style="crypto")
            self._fill_cell(sheet, row_index, 6, balance.final_balance, visual_style="bold", data_style="crypto")
            value = totals.setdefault(balance.holder, _ZERO)
            value += balance.final_balance
            totals[balance.holder] = value
            row_index += 1

        holder: str
        border_drawn: bool = False
        for holder, value in sorted(totals.items()):
            border_suffix: str = ""
            if not border_drawn:
                border_suffix = "_border"
                border_drawn = True
            self._fill_cell(sheet, row_index, 0, _("Total"), visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 1, holder, visual_style="bold" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 2, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 3, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 4, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 5, "", visual_style="transparent" + border_suffix, data_style="default")
            self._fill_cell(sheet, row_index, 6, value, visual_style="bold" + border_suffix, data_style="crypto")
            row_index += 1

        return row_index

    def __generate_average_price_per_unit(self, sheet: Any, asset: str, price_per_unit: RP2Decimal, row_index: int) -> int:
        self._fill_cell(sheet, row_index, 0, _("Average Price"), visual_style="title")
        self._fill_cell(sheet, row_index + 1, 0, _("Average Price"), visual_style="header")
        self._fill_cell(sheet, row_index + 2, 0, _("Paid Per 1 {}").format(asset), visual_style="header")
        self._fill_cell(sheet, row_index + 3, 0, price_per_unit, visual_style="transparent", data_style="fiat")

        return row_index + 4

    def __get_hyperlinked_transaction_value(self, transaction: AbstractTransaction, value: Any) -> Any:
        row: Optional[int] = self.__get_in_out_sheet_row(transaction)
        if not row:
            # This may occur if command line time filters are activated
            return value
        if isinstance(value, (RP2Decimal, int, float)):
            return f'=HYPERLINK("#{self.get_in_out_sheet_name(transaction.asset)}.a{row}:z{row}"; {value})'
        return f'=HYPERLINK("#{self.get_in_out_sheet_name(transaction.asset)}.a{row}:z{row}"; "{value}")'

    def __get_hyperlinked_summary_value(self, asset: str, value: Any, year: int) -> Any:
        row: int = self.__tax_sheet_year_2_row[_AssetAndYear(asset, year)]
        if isinstance(value, (RP2Decimal, int, float)):
            return f'=HYPERLINK("#{self.get_tax_sheet_name(asset)}.a{row}:z{row}"; {value})'
        return f'=HYPERLINK("#{self.get_tax_sheet_name(asset)}.a{row}:z{row}"; "{value}")'

    def __get_in_out_sheet_row(self, transaction: AbstractTransaction) -> Optional[int]:
        if transaction not in self.__in_out_sheet_transaction_2_row:
            return None
        return self.__in_out_sheet_transaction_2_row[transaction]

    def __generate_gain_loss_detail(self, sheet: Any, asset: str, computed_data: ComputedData, row_index: int) -> int:
        row_index = self._fill_header(
            _("Gain / Loss Detail"), self.__gain_loss_detail_header_names_row_1, self.__gain_loss_detail_header_names_row_2, sheet, row_index, 0
        )

        gain_loss_set: GainLossSet = computed_data.gain_loss_set

        taxable_event_style_modifier: str = ""
        acquired_lot_style_modifier: str = "_alt"
        year: int = 0
        border_style: _BorderStyle

        previous_acquired_lot: Optional[InTransaction] = None
        for entry in gain_loss_set:
            gain_loss: GainLoss = cast(GainLoss, entry)
            border_suffix: str = ""
            border_style = self.__get_border_style(gain_loss.taxable_event.timestamp.year, year)
            if gain_loss.taxable_event.timestamp.year != year:
                self.__tax_sheet_year_2_row[_AssetAndYear(asset, gain_loss.taxable_event.timestamp.year)] = row_index + 1
            year = border_style.year
            border_suffix = border_style.border_suffix
            transparent_style: str = f"transparent{border_suffix}"
            taxable_event_style: str = f"taxable_event{taxable_event_style_modifier}{border_suffix}"
            highlighted_style: str = f"highlighted{border_suffix}"
            current_taxable_event_fraction: int = gain_loss_set.get_taxable_event_fraction(gain_loss) + 1
            total_taxable_event_fractions: int = gain_loss_set.get_taxable_event_number_of_fractions(gain_loss.taxable_event)
            transaction_type: str = (
                f"{self._get_table_type_from_transaction(gain_loss.taxable_event)} / " f"{gain_loss.taxable_event.transaction_type.get_translation().upper()}"
            )
            taxable_event_note: str = (
                f"{current_taxable_event_fraction}/"
                f"{total_taxable_event_fractions}: "
                f"{gain_loss.crypto_amount:.8f} of "
                f"{gain_loss.taxable_event.crypto_balance_change:.8f} "
                f"{asset}"
            )
            acquired_lot_style: str

            self._fill_cell(sheet, row_index, 0, gain_loss.crypto_amount, visual_style=transparent_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 1, gain_loss.asset, visual_style=transparent_style)
            self._fill_cell(sheet, row_index, 2, computed_data.get_crypto_gain_loss_running_sum(gain_loss), visual_style=transparent_style, data_style="crypto")
            self._fill_cell(sheet, row_index, 3, gain_loss.fiat_gain, visual_style=transparent_style, data_style="fiat")
            self._fill_cell(sheet, row_index, 4, _("LONG") if gain_loss.is_long_term_capital_gains() else _("SHORT"), visual_style=transparent_style)
            self._fill_cell(
                sheet,
                row_index,
                5,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event.timestamp),
                visual_style=taxable_event_style,
            )
            self._fill_cell(
                sheet, row_index, 6, self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, transaction_type), visual_style=taxable_event_style
            )
            self._fill_cell(
                sheet,
                row_index,
                7,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event_fraction_percentage),
                visual_style=taxable_event_style,
                data_style="percent",
            )
            self._fill_cell(
                sheet,
                row_index,
                8,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event_fiat_amount_with_fee_fraction),
                visual_style=highlighted_style,
                data_style="fiat",
            )
            self._fill_cell(
                sheet,
                row_index,
                9,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event.spot_price),
                visual_style=taxable_event_style,
                data_style="fiat",
            )
            self._fill_cell(
                sheet,
                row_index,
                10,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, gain_loss.taxable_event.unique_id),
                visual_style=taxable_event_style,
                data_style="fiat",
            )
            self._fill_cell(
                sheet,
                row_index,
                11,
                self.__get_hyperlinked_transaction_value(gain_loss.taxable_event, taxable_event_note),
                visual_style=f"taxable_event_note{border_suffix}",
            )
            if current_taxable_event_fraction == total_taxable_event_fractions:
                # Last fraction: change color
                taxable_event_style_modifier = "" if taxable_event_style_modifier == "_alt" else "_alt"

            if gain_loss.acquired_lot:
                if gain_loss.acquired_lot != previous_acquired_lot:
                    # Last fraction: change color
                    acquired_lot_style_modifier = "" if acquired_lot_style_modifier == "_alt" else "_alt"
                acquired_lot_style = f"acquired_lot{acquired_lot_style_modifier}{border_suffix}"
                current_acquired_lot_fraction: int = gain_loss_set.get_acquired_lot_fraction(gain_loss) + 1
                total_acquired_lot_fractions: int = gain_loss_set.get_acquired_lot_number_of_fractions(gain_loss.acquired_lot)
                acquired_lot_note: str = (
                    f"{current_acquired_lot_fraction}/"
                    f"{total_acquired_lot_fractions}: "
                    f"{gain_loss.crypto_amount:.8f} of "
                    f"{gain_loss.acquired_lot.crypto_balance_change:.8f} "
                    f"{asset}"
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    12,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot.timestamp),
                    visual_style=acquired_lot_style,
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    13,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot_fraction_percentage),
                    visual_style=acquired_lot_style,
                    data_style="percent",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    14,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot_fiat_amount_with_fee_fraction),
                    visual_style=acquired_lot_style,
                    data_style="fiat",
                )
                fiat_fee_fraction: RP2Decimal = gain_loss.acquired_lot.fiat_fee * gain_loss.acquired_lot_fraction_percentage
                self._fill_cell(
                    sheet,
                    row_index,
                    15,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, fiat_fee_fraction),
                    visual_style=acquired_lot_style,
                    data_style="fiat",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    16,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.fiat_cost_basis),
                    visual_style=highlighted_style,
                    data_style="fiat",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    17,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot.spot_price),
                    visual_style=acquired_lot_style,
                    data_style="fiat",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    18,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, gain_loss.acquired_lot.unique_id),
                    visual_style=acquired_lot_style,
                    data_style="fiat",
                )
                self._fill_cell(
                    sheet,
                    row_index,
                    19,
                    self.__get_hyperlinked_transaction_value(gain_loss.acquired_lot, acquired_lot_note),
                    visual_style=f"acquired_lot_note{border_suffix}",
                )

                previous_acquired_lot = gain_loss.acquired_lot
            else:
                acquired_lot_style = f"acquired_lot{acquired_lot_style_modifier}{border_suffix}"
                for i in range(12, 19):
                    self._fill_cell(sheet, row_index, i, "", visual_style=f"{acquired_lot_style}")

            row_index += 1

        return row_index

    def __generate_yearly_gain_loss_summary(self, sheet: Any, asset: str, yearly_gain_loss_list: List[YearlyGainLoss], row_index: int) -> int:
        for gain_loss in yearly_gain_loss_list:
            visual_style: str = "transparent"
            capital_gains_type: str = _("LONG") if gain_loss.is_long_term_capital_gains else _("SHORT")
            year: int = gain_loss.year
            self._fill_cell(sheet, row_index, 0, self.__get_hyperlinked_summary_value(asset, year, year), visual_style=visual_style)
            self._fill_cell(sheet, row_index, 1, self.__get_hyperlinked_summary_value(asset, asset, year), visual_style=visual_style)
            self._fill_cell(
                sheet,
                row_index,
                2,
                self.__get_hyperlinked_summary_value(asset, gain_loss.fiat_gain_loss, year),
                visual_style=visual_style,
                data_style="fiat",
            )
            self._fill_cell(sheet, row_index, 3, self.__get_hyperlinked_summary_value(asset, capital_gains_type, year), visual_style=visual_style)
            self._fill_cell(
                sheet,
                row_index,
                4,
                self.__get_hyperlinked_summary_value(asset, gain_loss.transaction_type.get_translation().upper(), year),
                visual_style=visual_style,
            )
            self._fill_cell(
                sheet, row_index, 5, self.__get_hyperlinked_summary_value(asset, gain_loss.crypto_amount, year), visual_style=visual_style, data_style="crypto"
            )
            self._fill_cell(
                sheet,
                row_index,
                6,
                self.__get_hyperlinked_summary_value(asset, gain_loss.fiat_amount, year),
                visual_style=visual_style,
                data_style="fiat",
            )
            self._fill_cell(
                sheet,
                row_index,
                7,
                self.__get_hyperlinked_summary_value(asset, gain_loss.fiat_cost_basis, year),
                visual_style=visual_style,
                data_style="fiat",
            )
            row_index += 1

        return row_index



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Optional

import ezodf

from rp2.abstract_transaction import AbstractTransaction
from rp2.configuration import MAX_DATE, MIN_DATE, Configuration
from rp2.entry_types import EntrySetType, TransactionType
from rp2.in_transaction import InTransaction
from rp2.input_data import InputData
from rp2.intra_transaction import IntraTransaction
from rp2.logger import LOGGER
from rp2.out_transaction import OutTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2Error, RP2RuntimeError, RP2ValueError
from rp2.transaction_set import TransactionSet

_TABLE_END: str = "TABLE END"


def open_ods(configuration: Configuration, input_file_path: str) -> Any:
    Configuration.type_check("configuration", configuration)
    configuration.type_check_string("input_file_path", input_file_path)

    if not Path(input_file_path).exists():
        raise RP2ValueError(f"Error: {input_file_path} does not exist")

    return ezodf.opendoc(input_file_path)


def parse_ods(configuration: Configuration, asset: str, input_file_handle: Any) -> InputData:  # pylint: disable=too-many-branches
    Configuration.type_check("configuration", configuration)
    configuration.type_check_asset("asset", asset)

    if asset not in input_file_handle.sheets.names():
        raise RP2ValueError(f"Error: sheet {asset} does not exist in {Path(input_file_handle.docname).resolve()}")
    input_sheet: Any = input_file_handle.sheets[asset]

    unfiltered_transaction_sets: Dict[EntrySetType, TransactionSet] = {}

    unfiltered_transaction_sets[EntrySetType.IN] = TransactionSet(configuration, "IN", asset, MIN_DATE, MAX_DATE)
    unfiltered_transaction_sets[EntrySetType.OUT] = TransactionSet(configuration, "OUT", asset, MIN_DATE, MAX_DATE)
    unfiltered_transaction_sets[EntrySetType.INTRA] = TransactionSet(configuration, "INTRA", asset, MIN_DATE, MAX_DATE)

    artificial_transaction_list: List[AbstractTransaction] = []

    current_table_type: Optional[EntrySetType] = None
    current_table_row_count: int = 0
    i: int = 0
    row: Any = None
    # Used for artificial transactions only: e.g. the fee-only transaction that is created artificially to model crypto fee of in-transactions.
    # Artificial internal ids are negative.
    artificial_internal_id = 0
    for i, row in enumerate(input_sheet.rows()):
        artificial_internal_id -= 1
        cell0_value: str = row[0].value
        # The numeric elements of the row_values list are used to initialize RP2Decimal instances. In theory we could collect string representations
        # from numeric strings using the plaintext() method of Cell, but this doesn't work well because of an ezodf limitation: such strings are
        # affected by the format of their cell (so they may be less precise than their real value, depending on cell format), so as a workaround
        # we read the value attribute which returns a float. In theory this could cause precision loss, but in reality most of the input data from
        # wallets / exchanges is very low-precision, so using a float is almost always adequate. The only exception would be if a wallet / exchange
        # input data had numbers with more than CRYPTO_DECIMALS (defined in rp2_decimal.py) decimal digits, which is quite uncommon: in this case
        # RP2 would still work, but it would have a little precision loss on these high-precision numbers. Also read the comments in
        # _process_constructor_argument_pack().
        row_values: List[Any] = [cell.value for cell in row]
        LOGGER.debug("parsing row: %s", row_values)

        if current_table_type is not None:
            # Inside a table
            if _is_table_begin(cell0_value):
                # Found a nested table begin
                raise RP2ValueError(f'{asset}({i + 1}): Found "{cell0_value}" keyword while parsing table {current_table_type}')
            if _is_empty(cell0_value):
                # Found an empty cell inside a table
                raise RP2ValueError(f"{asset}({i + 1}): Found an empty cell while parsing table {current_table_type}")

        else:
            # Outside a table
            if _is_table_end(cell0_value):
                # Found a spurious table end
                raise RP2ValueError(f"{asset}({i + 1}): Found end-table keyword without having found a table-begin keyword first")
            if not _is_empty(cell0_value) and not _is_table_begin(cell0_value):
                # Found a non-empty and non-table-begin cell outside a table
                raise RP2ValueError(f'{asset}({i + 1}): Found an invalid cell "{cell0_value}" while looking for a table-begin token')

        if _is_table_begin(cell0_value):
            # New table start
            current_table_row_count = 0
            current_table_type = _get_entry_set_type(cell0_value)
            if current_table_type and not unfiltered_transaction_sets[current_table_type].is_empty():
                # Found an already-processed table type
                raise RP2ValueError(f"{asset}({i + 1}): Found more than one {cell0_value} symbol")
        elif _is_table_end(cell0_value):
            # Table end
            current_table_type = None
        elif current_table_type is not None and current_table_row_count == 1:
            # Header line: make sure it's not transaction data
            try:
                _create_transaction(configuration, current_table_type, i + 1, row_values)
            except Exception:  # pylint: disable=broad-except  # nosec
                # Couldn't create transaction as expected: this is a table header
                # TODO: this could still be a transaction but with some bad fields that would  # pylint: disable=fixme
                # cause an exception. In this case this logic would incorrectly assume it's a
                # header. Can we do better in this case? Some heuristics testing field by
                # field might help.
                pass
            else:
                raise RP2ValueError(f"{asset}({i + 1}): Found data with no header")
        elif current_table_type is not None and current_table_row_count > 1:
            # Transaction line
            _create_and_process_transaction(
                configuration, row_values, current_table_type, i + 1, artificial_internal_id, unfiltered_transaction_sets, artificial_transaction_list
            )
        current_table_row_count += 1

    if current_table_type is not None:
        raise RP2ValueError(f"TABLE END not found for {current_table_type} table")
    if unfiltered_transaction_sets[EntrySetType.IN].is_empty():
        raise RP2ValueError(f"{asset}: IN table not found or empty")

    for transaction in artificial_transaction_list:
        if isinstance(transaction, InTransaction):
            unfiltered_transaction_sets[EntrySetType.IN].add_entry(transaction)
        elif isinstance(transaction, OutTransaction):
            unfiltered_transaction_sets[EntrySetType.OUT].add_entry(transaction)
        elif isinstance(transaction, IntraTransaction):
            unfiltered_transaction_sets[EntrySetType.INTRA].add_entry(transaction)
        else:
            raise RP2ValueError(f"Internal error: invalid transaction class: {transaction}")

    return InputData(
        asset,
        unfiltered_transaction_sets[EntrySetType.IN],
        unfiltered_transaction_sets[EntrySetType.OUT],
        unfiltered_transaction_sets[EntrySetType.INTRA],
        configuration.from_date,
        configuration.to_date,
    )


def _create_and_process_transaction(
    configuration: Configuration,
    row_values: List[Any],
    current_table_type: EntrySetType,
    internal_id: int,
    artificial_internal_id: int,
    unfiltered_transaction_sets: Dict[EntrySetType, TransactionSet],
    artificial_transaction_list: List[AbstractTransaction],
) -> None:
    transaction: AbstractTransaction = _create_transaction(configuration, current_table_type, internal_id, row_values)

    if isinstance(transaction, InTransaction) and transaction.is_crypto_fee_defined:
        # If an InTransaction has crypto fee defined it is split into two transactions:
        # - InTransaction with crypto_fee set to 0, but fiat_fee left as-is (the fiat-converted value of crypto_fee),
        # - fee-typed OutTransaction, modeling the crypto_fee.
        # These two transactions correctly model the coin flow of a InTransaction with crypto fee > 0: their notes
        # fields are updated with a description of the above.
        notes: str = f"{transaction.notes}; " if transaction.notes else ""
        notes = (
            f"{notes}This transaction has a crypto fee of {transaction.crypto_fee} {transaction.asset}, "
            "which is modeled with an artificial, fee-only out-transaction (look for it among out-transactions)"
        )

        unfiltered_transaction_sets[EntrySetType.IN].add_entry(
            InTransaction(
                configuration=configuration,
                timestamp=f"{transaction.timestamp}",
                asset=transaction.asset,
                exchange=transaction.exchange,
                holder=transaction.holder,
                transaction_type=transaction.transaction_type.value,
                spot_price=transaction.spot_price,
                crypto_in=transaction.crypto_in,
                crypto_fee=None,
                fiat_in_no_fee=transaction.fiat_in_no_fee,
                fiat_in_with_fee=transaction.fiat_in_with_fee,
                fiat_fee=transaction.fiat_fee,
                internal_id=internal_id,
                unique_id=transaction.unique_id,
                notes=notes,
            )
        )
        artificial_transaction_list.append(
            OutTransaction(
                configuration=configuration,
                timestamp=f"{transaction.timestamp}",
                asset=transaction.asset,
                exchange=transaction.exchange,
                holder=transaction.holder,
                transaction_type=TransactionType.FEE.value,
                spot_price=transaction.spot_price,
                crypto_out_no_fee=ZERO,
                crypto_fee=transaction.crypto_fee,
                internal_id=artificial_internal_id,
                unique_id=transaction.unique_id,
                notes=(
                    f"Artificial transaction modeling the crypto fee of {transaction.crypto_fee} {transaction.asset} "
                    f"of the in-transaction that occurred on {transaction.timestamp} (look for it among in-transactions)"
                ),
            )
        )
    else:
        unfiltered_transaction_sets[current_table_type].add_entry(transaction)


# Returns all numeric parameters of the constructor: used in construction of __init__ argument pack to parse such parameters as decimals
@lru_cache(maxsize=None, typed=False)
def _get_decimal_constructor_argument_names(class_name: str) -> List[str]:
    result: List[str] = []
    class_to_inspect: Any
    if class_name not in globals():
        raise RP2RuntimeError(f"Internal error: couldn't find class {class_name}")
    class_to_inspect = globals()[class_name]
    if not issubclass(class_to_inspect, AbstractTransaction):
        raise RP2RuntimeError(f"Internal error: class {class_name} is not a subclass of AbstractTransaction")
    arg_spec = inspect.getfullargspec(class_to_inspect.__init__)
    for parameter_name, parameter_type in arg_spec.annotations.items():
        if parameter_type in [RP2Decimal, Optional[RP2Decimal]]:
            result.append(parameter_name)
    return result


# Add configuration and internal_id to argument pack and turn floats to strings to maximize decimal precision. See comment inside the function.
def _process_constructor_argument_pack(
    configuration: Configuration,
    argument_pack: Dict[str, Any],
    internal_id: int,
    class_name: str,
) -> Dict[str, Any]:
    argument_pack.update({"configuration": configuration, "internal_id": internal_id})
    numeric_parameters: List[str] = _get_decimal_constructor_argument_names(class_name)
    for numeric_parameter in numeric_parameters:
        if numeric_parameter in argument_pack:
            try:
                value = argument_pack[numeric_parameter]
                if value == "__unknown":
                    # If value is __unknown, this transaction has been generated by DaLI and it is unresolved
                    del argument_pack["configuration"]
                    raise RP2RuntimeError(
                        f"Encountered an unresolved DaLI transaction (read DaLI's documentation / FAQ to learn how to resolve this issue): {argument_pack}"
                    )

                # It would be ideal to pass a string directly to the RP2Decimal constructor for maximum precision, but due to ezodf limitations we
                # cannot get the string representation directly from the spreadsheet (see the comment on cell format inside parse_ods() for more
                # detail), so at parse time we have to get the float value from the cell. Here we convert the float to string, which allows us to
                # initialize a maximum-precision RP2Decimal (11 decimal digits is enough precision for millisats).
                argument_pack[numeric_parameter] = RP2Decimal(f"{value:.11f}") if value is not None else None
            except (ValueError, RP2Error) as exc:
                raise RP2ValueError(f"Argument '{numeric_parameter}' has non-numeric value: {value}") from exc

    return argument_pack


def _create_transaction(
    configuration: Configuration,
    entry_set_type: EntrySetType,
    internal_id: int,
    row_values: List[Any],
) -> AbstractTransaction:
    transaction: AbstractTransaction
    EntrySetType.type_check("entry_set_type", entry_set_type)
    configuration.type_check_internal_id("internal_id", internal_id)
    if entry_set_type == EntrySetType.IN:
        argument_pack: Dict[str, Any] = configuration.get_in_table_constructor_argument_pack(row_values)
        argument_pack = _process_constructor_argument_pack(configuration, argument_pack, internal_id, "InTransaction")
        transaction = InTransaction(**argument_pack)
    elif entry_set_type == EntrySetType.OUT:
        argument_pack = configuration.get_out_table_constructor_argument_pack(row_values)
        argument_pack = _process_constructor_argument_pack(configuration, argument_pack, internal_id, "OutTransaction")
        transaction = OutTransaction(**argument_pack)
    elif entry_set_type == EntrySetType.INTRA:
        argument_pack = configuration.get_intra_table_constructor_argument_pack(row_values)
        argument_pack = _process_constructor_argument_pack(configuration, argument_pack, internal_id, "IntraTransaction")
        transaction = IntraTransaction(**argument_pack)
    return transaction


def _get_entry_set_type(cell_value: str) -> Optional[EntrySetType]:
    return EntrySetType.get_entry_set_type_from_string(cell_value)


def _is_table_in(cell_value: str) -> bool:
    return _get_entry_set_type(cell_value) == EntrySetType.IN


def _is_table_out(cell_value: str) -> bool:
    return _get_entry_set_type(cell_value) == EntrySetType.OUT


def _is_table_intra(cell_value: str) -> bool:
    return _get_entry_set_type(cell_value) == EntrySetType.INTRA


def _is_table_begin(cell_value: str) -> bool:
    return _is_table_in(cell_value) or _is_table_out(cell_value) or _is_table_intra(cell_value)


def _is_table_end(cell_value: str) -> bool:
    return cell_value == _TABLE_END


def _is_empty(cell_value: str) -> bool:
    return cell_value is None or cell_value == ""


def main() -> None:
    pass


if __name__ == "__main__":
    main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from typing import List, Optional, cast

from dateutil.parser import parse

from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.configuration import Configuration
from rp2.entry_types import EntrySetType, TransactionType
from rp2.in_transaction import InTransaction
from rp2.intra_transaction import IntraTransaction
from rp2.out_transaction import OutTransaction
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError
from rp2.transaction_set import TransactionSet


class TestTransactionSet(unittest.TestCase):
    _configuration: Configuration

    @classmethod
    def setUpClass(cls) -> None:
        TestTransactionSet._configuration = Configuration("./config/test_data.ini", US())

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_good_transaction_set(self) -> None:
        transaction_set: TransactionSet = TransactionSet(self._configuration, "MIXED", "B1")

        # Test empty iterator
        count: int = 0
        entry: AbstractEntry
        transaction: AbstractTransaction
        for entry in transaction_set:
            transaction = cast(AbstractTransaction, entry)
            count += 1
        self.assertEqual(count, 0)

        transaction3: InTransaction = InTransaction(
            self._configuration,
            "1/8/2021 8:42:43.883 -04:00",
            "B1",
            "Coinbase",
            "Alice",
            "bUy",
            RP2Decimal("1000"),
            RP2Decimal("3.0002"),
            fiat_fee=RP2Decimal("20"),
            fiat_in_no_fee=RP2Decimal("3000.2"),
            fiat_in_with_fee=RP2Decimal("3020.2"),
            internal_id=30,
        )
        transaction_set.add_entry(transaction3)

        transaction2: InTransaction = InTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "BlockFi",
            "Bob",
            "InTeREST",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("0"),
            fiat_in_no_fee=RP2Decimal("2000.2"),
            fiat_in_with_fee=RP2Decimal("2000.2"),
            internal_id=20,
        )
        transaction_set.add_entry(transaction2)

        timestamps: List[str] = [
            "2021-01-02T08:42:43.882Z",
            "1/8/2021 8:42:43.883 -04:00",
        ]

        count = 0
        timestamp: str
        for transaction, timestamp in zip(transaction_set, timestamps):  # type: ignore
            self.assertEqual(transaction.timestamp, parse(timestamp))
            count += 1
        self.assertEqual(count, 2)

        transaction1: OutTransaction = OutTransaction(
            self._configuration,
            "6/1/2020 3:59:59 -04:00",
            "B1",
            "Coinbase Pro",
            "Bob",
            "SeLL",
            RP2Decimal("900.9"),
            RP2Decimal("2.2"),
            RP2Decimal("0"),
            internal_id=10,
        )
        transaction_set.add_entry(transaction1)

        transaction5: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-04-02T08:42:43.882Z",
            "B1",
            "Coinbase Pro",
            "Bob",
            "BlockFi",
            "Alice",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            RP2Decimal("1.9998"),
            internal_id=50,
        )
        transaction_set.add_entry(transaction5)

        transaction4: IntraTransaction = IntraTransaction(
            self._configuration,
            "2021-03-28T08:42:43.882Z",
            "B1",
            "Coinbase",
            "Bob",
            "Coinbase",
            "Alice",
            RP2Decimal("100.0"),
            RP2Decimal("30"),
            RP2Decimal("30"),
            internal_id=40,
        )
        transaction_set.add_entry(transaction4)

        self.assertEqual(transaction_set.entry_set_type, EntrySetType.MIXED)
        self.assertEqual(transaction_set.asset, "B1")

        internal_ids: List[str] = ["10", "20", "30", "40", "50"]
        transactions: List[AbstractTransaction] = [transaction1, transaction2, transaction3, transaction4, transaction5]
        parents: List[Optional[AbstractTransaction]] = [
            None,
            transaction1,
            transaction2,
            transaction3,
            transaction4,
        ]
        timestamps = [
            "6/1/2020 3:59:59 -04:00",
            "2021-01-02T08:42:43.882Z",
            "1/8/2021 8:42:43.883 -04:00",
            "2021-03-28T08:42:43.882Z",
            "2021-04-02T08:42:43.882Z",
        ]
        transaction_types: List[TransactionType] = [
            TransactionType.SELL,
            TransactionType.INTEREST,
            TransactionType.BUY,
            TransactionType.MOVE,
            TransactionType.MOVE,
        ]
        fiat_taxable_amounts: List[RP2Decimal] = [RP2Decimal(s) for s in ["1981.98", "2000.2", "0", "0", "0.4"]]
        crypto_balance_changes: List[RP2Decimal] = [RP2Decimal(s) for s in ["2.2", "2.0002", "3.0002", "0", "0.0004"]]
        fiat_balance_changes: List[RP2Decimal] = [RP2Decimal(s) for s in ["1981.98", "2000.2", "3020.2", "0", "0.4"]]

        count = 0
        expected_transaction: AbstractTransaction
        parent: AbstractTransaction
        internal_id: str
        transaction_type: TransactionType
        fiat_taxable_amount: RP2Decimal
        crypto_balance_change: RP2Decimal
        fiat_balance_change: RP2Decimal
        for (  # type: ignore
            transaction,
            expected_transaction,
            parent,
            internal_id,
            timestamp,
            transaction_type,
            fiat_taxable_amount,
            crypto_balance_change,
            fiat_balance_change,
        ) in zip(  # type: ignore
            transaction_set,
            transactions,
            parents,
            internal_ids,
            timestamps,
            transaction_types,
            fiat_taxable_amounts,
            crypto_balance_changes,
            fiat_balance_changes,
        ):
            self.assertEqual(transaction, expected_transaction)
            self.assertEqual(transaction_set.get_parent(transaction), parent)
            self.assertEqual(transaction.internal_id, internal_id)
            self.assertEqual(transaction.timestamp, parse(timestamp))
            self.assertEqual(transaction.transaction_type, transaction_type)
            self.assertEqual(transaction.asset, "B1")
            self.assertEqual(transaction.fiat_taxable_amount, fiat_taxable_amount)
            self.assertEqual(transaction.crypto_balance_change, crypto_balance_change)
            self.assertEqual(transaction.fiat_balance_change, fiat_balance_change)
            count += 1
        self.assertEqual(count, 5)

        self.assertTrue(str(transaction_set).startswith("TransactionSet:\n  configuration=./config/test_data.ini\n  entry_set_type=EntrySetType.MIXED"))

    def test_bad_transaction_set(self) -> None:
        in_transaction = InTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "BlockFi",
            "Bob",
            "INTERest",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("0"),
            fiat_in_with_fee=RP2Decimal("2000.2"),
            fiat_in_no_fee=RP2Decimal("2000.2"),
            internal_id=20,
        )
        # Different instance with same contents as in_transaction
        in_transaction2 = InTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "BlockFi",
            "Bob",
            "interEST",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("0"),
            fiat_in_with_fee=RP2Decimal("2000.2"),
            fiat_in_no_fee=RP2Decimal("2000.2"),
            internal_id=20,
        )
        out_transaction = OutTransaction(
            self._configuration,
            "6/1/2020 3:59:59 -04:00",
            "B1",
            "Coinbase Pro",
            "Bob",
            "sELl",
            RP2Decimal("900.9"),
            RP2Decimal("2.2"),
            RP2Decimal("0"),
            internal_id=10,
        )
        intra_transaction = IntraTransaction(
            self._configuration,
            "2021-04-02T08:42:43.882Z",
            "B1",
            "Coinbase Pro",
            "Bob",
            "BlockFi",
            "Alice",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            RP2Decimal("1.9998"),
            internal_id=50,
        )

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            TransactionSet(None, "IN", "B1")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            TransactionSet(1111, "IN", "B1")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry_set_type' has non-string value .*"):
            # Bad transaction set type
            TransactionSet(self._configuration, None, "B1")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter .* has invalid entry set type value: .*"):
            # Bad transaction set type
            TransactionSet(self._configuration, "foobar", "B1")
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry_set_type' has non-string value .*"):
            # Bad transaction set type
            TransactionSet(self._configuration, 1111, "B1")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            # Bad asset
            TransactionSet(self._configuration, "IN", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            # Bad asset
            TransactionSet(self._configuration, "IN", 1111)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            # Bad asset
            TransactionSet(self._configuration, "IN", "Qwerty")
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_date' is not of type date"):
            # Bad from_date
            TransactionSet(self._configuration, "IN", "B1", "2018")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_date' is not of type date"):
            # Bad from_date
            TransactionSet(self._configuration, "IN", "B1", "foobar")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_date' is not of type date"):
            # Bad to_date
            TransactionSet(self._configuration, "IN", "B1", "2018-01-03", "2019-07")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_date' is not of type date"):
            # Bad to_date
            TransactionSet(self._configuration, "IN", "B1", "2018-01-03", 0)  # type: ignore

        in_transaction_set = TransactionSet(self._configuration, "IN", "B1")
        out_transaction_set = TransactionSet(self._configuration, "OUT", "B1")
        intra_transaction_set = TransactionSet(self._configuration, "INTRA", "B1")

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type AbstractTransaction: .*"):
            # Bad transaction add
            in_transaction_set.add_entry(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type AbstractTransaction: .*"):
            # Bad transaction add
            in_transaction_set.add_entry(1111)  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Attempting to add a .* to a set of type IN"):
            # Transaction add type mismatch
            in_transaction_set.add_entry(out_transaction)
        with self.assertRaisesRegex(RP2TypeError, "Attempting to add a .* to a set of type OUT"):
            # Transaction add type mismatch
            out_transaction_set.add_entry(intra_transaction)
        with self.assertRaisesRegex(RP2TypeError, "Attempting to add a .* to a set of type INTRA"):
            # Transaction add type mismatch
            intra_transaction_set.add_entry(in_transaction)

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type AbstractTransaction: .*"):
            # Bad get_parent parameter
            in_transaction_set.get_parent(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'entry' is not of type AbstractTransaction: .*"):
            # Bad get_parent parameter
            in_transaction_set.get_parent(1111)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Unknown entry:.*"):
            # Unknown get_parent parameter
            in_transaction_set.get_parent(out_transaction)
        with self.assertRaisesRegex(RP2ValueError, "Unknown entry:.*"):
            # Unknown get_parent parameter
            out_transaction_set.get_parent(intra_transaction)
        with self.assertRaisesRegex(RP2ValueError, "Unknown entry:.*"):
            # Unknown get_parent parameter
            intra_transaction_set.get_parent(in_transaction)
        with self.assertRaisesRegex(RP2ValueError, "Entry already added: InTransaction"):
            in_transaction_set = TransactionSet(self._configuration, "IN", "B1")
            in_transaction_set.add_entry(in_transaction)
            in_transaction_set.add_entry(in_transaction)
        with self.assertRaisesRegex(RP2ValueError, "Entry already added: InTransaction"):
            in_transaction_set = TransactionSet(self._configuration, "IN", "B1")
            in_transaction_set.add_entry(in_transaction)
            in_transaction_set.add_entry(in_transaction2)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional

from rp2.abstract_transaction import AbstractTransaction
from rp2.configuration import Configuration
from rp2.logger import LOGGER
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


class IntraTransaction(AbstractTransaction):
    def __init__(
        self,
        configuration: Configuration,
        timestamp: str,
        asset: str,
        from_exchange: str,
        from_holder: str,
        to_exchange: str,
        to_holder: str,
        spot_price: Optional[RP2Decimal],
        crypto_sent: RP2Decimal,
        crypto_received: RP2Decimal,
        internal_id: Optional[int] = None,
        unique_id: Optional[str] = None,
        notes: Optional[str] = None,
    ) -> None:
        Configuration.type_check("configuration", configuration)
        self.__crypto_sent: RP2Decimal = configuration.type_check_positive_decimal("crypto_sent", crypto_sent, non_zero=True)
        self.__crypto_received: RP2Decimal = configuration.type_check_positive_decimal("crypto_received", crypto_received)
        self.__crypto_fee: RP2Decimal = self.__crypto_sent - self.__crypto_received
        if spot_price is None or (isinstance(spot_price, RP2Decimal) and spot_price == ZERO):
            # Sometimes, when fee is 0 in IntraTransactions, exchanges don't provide the spot_price:
            # - if the fee is 0, this is OK because spot price isn't needed (in this case spot price is assigned 0).
            # - if the fee is 0, raise an exception
            if self.__crypto_fee == ZERO:
                spot_price = ZERO
            else:
                raise RP2ValueError(
                    f"crypto_fee is non-zero ({self.__crypto_fee}) but spot_price is empty or zero: {timestamp} {asset} {crypto_sent} {unique_id} "
                )
        super().__init__(configuration, timestamp, asset, "MOVE", spot_price, internal_id, unique_id, notes)

        self.__from_exchange: str = configuration.type_check_exchange("from_exchange", from_exchange)
        self.__from_holder: str = configuration.type_check_holder("from_holder", from_holder)
        self.__to_exchange: str = configuration.type_check_exchange("to_exchange", to_exchange)
        self.__to_holder: str = configuration.type_check_holder("to_holder", to_holder)
        self.__fiat_fee: RP2Decimal

        if self.__from_exchange == self.__to_exchange and self.__from_holder == self.__to_holder:
            LOGGER.warning(
                "%s %s (%s, id %s): from/to exchanges/holders are the same: sending to self",
                self.asset,
                type(self).__name__,
                str(self.timestamp),
                self.internal_id,
            )
        if self.__crypto_sent < self.__crypto_received:
            raise RP2ValueError(f"{self.asset} {type(self).__name__} ({self.timestamp}, id {self.internal_id}): crypto sent < crypto received")

        self.__fiat_fee = self.__crypto_fee * self.spot_price

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        self.configuration.type_check_positive_int("indent", indent)
        self.configuration.type_check_bool("repr_format", repr_format)
        if extra_data and not isinstance(extra_data, List):
            raise RP2TypeError(f"Parameter 'extra_data' is not of type List: {extra_data}")

        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = repr
        if not repr_format:
            stringify = str
        class_specific_data = [
            f"from_exchange={stringify(self.from_exchange)}",
            f"from_holder={stringify(self.from_holder)}",
            f"to_exchange={stringify(self.to_exchange)}",
            f"to_holder={stringify(self.to_holder)}",
            f"transaction_type={stringify(self.transaction_type)}",
            f"spot_price={self.spot_price:.4f}",
            f"crypto_sent={self.crypto_sent:.8f}",
            f"crypto_received={self.crypto_received:.8f}",
            f"crypto_fee={self.crypto_fee:.8f}",
            f"fiat_fee={self.fiat_fee:.4f}",
            f"unique_id={self.unique_id}",
            f"is_taxable={stringify(self.is_taxable())}",
            f"fiat_taxable_amount={self.fiat_taxable_amount:.4f}",
        ]
        if extra_data:
            class_specific_data.extend(extra_data)

        return super().to_string(indent=indent, repr_format=repr_format, extra_data=class_specific_data)

    @property
    def from_exchange(self) -> str:
        return self.__from_exchange

    @property
    def from_holder(self) -> str:
        return self.__from_holder

    @property
    def to_exchange(self) -> str:
        return self.__to_exchange

    @property
    def to_holder(self) -> str:
        return self.__to_holder

    @property
    def crypto_sent(self) -> RP2Decimal:
        return self.__crypto_sent

    @property
    def crypto_received(self) -> RP2Decimal:
        return self.__crypto_received

    @property
    def crypto_fee(self) -> RP2Decimal:
        return self.__crypto_fee

    @property
    def fiat_fee(self) -> RP2Decimal:
        return self.__fiat_fee

    @property
    def crypto_taxable_amount(self) -> RP2Decimal:
        return self.crypto_fee

    @property
    def fiat_taxable_amount(self) -> RP2Decimal:
        return self.fiat_fee

    @property
    def crypto_deduction(self) -> RP2Decimal:
        return ZERO

    @property
    def fiat_deduction(self) -> RP2Decimal:
        return ZERO

    @property
    def crypto_balance_change(self) -> RP2Decimal:
        return self.crypto_fee

    @property
    def fiat_balance_change(self) -> RP2Decimal:
        return self.fiat_fee

    def is_taxable(self) -> bool:
        return self.fiat_fee > ZERO



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Set

from rp2.rp2_error import RP2TypeError


class AbstractCountry:
    @classmethod
    def type_check(cls, name: str, instance: "AbstractCountry") -> "AbstractCountry":
        if not isinstance(name, str):
            raise RP2TypeError(f"Parameter name is not a string: {repr(name)}")
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __init__(
        self,
        country_iso_code: str,  # ISO 3166-1 alpha-2
        currency_iso_code: str,  # ISO 4217
    ) -> None:
        if not isinstance(country_iso_code, str):
            raise RP2TypeError(f"Parameter 'country_iso_code' has non-string value {repr(country_iso_code)}")
        if not isinstance(currency_iso_code, str):
            raise RP2TypeError(f"Parameter 'currency_iso_code' has non-string value {repr(currency_iso_code)}")

        self.__country_iso_code = country_iso_code
        self.__currency_iso_code = currency_iso_code

    def __str__(self) -> str:
        output: List[str] = []
        output.append(f"{type(self).__name__}:")
        output.append(f"  country_iso_code={str(self.country_iso_code)}")
        output.append(f"  currency_iso_code={str(self.currency_iso_code)}")
        output.append(f"  long_term_capital_gain_period={str(self.get_long_term_capital_gain_period())}")
        return "\n".join(output)

    def __repr__(self) -> str:
        output: List[str] = []
        output.append(f"{type(self).__name__}(")
        output.append(f"country_iso_code={str(self.country_iso_code)}")
        output.append(f", currency_iso_code={str(self.currency_iso_code)}")
        output.append(f", long_term_capital_gain_period={str(self.get_long_term_capital_gain_period())}")
        output.append(")")
        return "".join(output)

    @property
    def country_iso_code(self) -> str:
        return self.__country_iso_code

    @property
    def currency_iso_code(self) -> str:
        return self.__currency_iso_code

    # Measured in days
    def get_long_term_capital_gain_period(self) -> int:
        raise NotImplementedError("Abstract function")

    # Default accounting method to use if the user doesn't specify one on the command line
    def get_default_accounting_method(self) -> str:
        raise NotImplementedError("Abstract function")

    # Set of accounting methods accepted in the country
    def get_accounting_methods(self) -> Set[str]:
        raise NotImplementedError("Abstract function")

    # Default set of generators to use if the user doesn't specify them on the command line
    def get_report_generators(self) -> Set[str]:
        raise NotImplementedError("Abstract function")

    # Default language to use at report generation if the user doesn't specify it on the command line (in ISO 639-1 format)
    def get_default_generation_language(self) -> str:
        raise NotImplementedError("Abstract function")



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Dict, Iterator, List, NamedTuple, Optional

from prezzemolo.avl_tree import AVLTree

from rp2.abstract_accounting_method import (
    AbstractAccountingMethod,
    AcquiredLotAndAmount,
    AcquiredLotCandidates,
)
from rp2.abstract_transaction import AbstractTransaction
from rp2.in_transaction import InTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError, RP2TypeError


class TaxableEventAndAcquiredLot(NamedTuple):
    taxable_event: AbstractTransaction
    acquired_lot: Optional[InTransaction]
    taxable_event_amount: RP2Decimal
    acquired_lot_amount: RP2Decimal


@dataclass(frozen=True, eq=True)
class _AcquiredLotAndIndex:
    acquired_lot: InTransaction
    index: int


class _LotExhaustedException(Exception):
    def __init__(self, message: str = "") -> None:
        self.__message = message
        super().__init__(self.__message)

    def __repr__(self) -> str:
        return self.message

    @property
    def message(self) -> str:
        return self.__message


class TaxableEventsExhaustedException(_LotExhaustedException):
    pass


class AcquiredLotsExhaustedException(_LotExhaustedException):
    pass


class AccountingEngine:
    __taxable_event_iterator: Iterator[AbstractTransaction]
    __acquired_lot_list: List[InTransaction]
    __acquired_lot_avl: AVLTree[str, _AcquiredLotAndIndex]
    __acquired_lot_2_partial_amount: Dict[InTransaction, RP2Decimal]

    # Disambiguation is needed for transactions that have the same timestamp, because the avl tree class expects unique keys: 12 decimal digits express
    # 1 quadrillion, which should be enough to capture the maximum number of same-timestamp transactions in all reasonable cases.
    KEY_DISAMBIGUATOR_LENGTH: int = 12
    MAX_KEY_DISAMBIGUATOR = "9" * KEY_DISAMBIGUATOR_LENGTH

    @classmethod
    def type_check(cls, name: str, instance: "AccountingEngine") -> "AccountingEngine":
        if not isinstance(name, str):
            raise RP2TypeError(f"Parameter name is not a string: {repr(name)}")
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __init__(self, years_2_methods: AVLTree[int, AbstractAccountingMethod]) -> None:
        self.__years_2_methods: AVLTree[int, AbstractAccountingMethod] = years_2_methods
        if not self.__years_2_methods:
            raise RP2RuntimeError("Internal error: no accounting method defined")

    # Iterators yield transactions in ascending chronological order
    def initialize(
        self,
        taxable_event_iterator: Iterator[AbstractTransaction],
        acquired_lot_iterator: Iterator[InTransaction],
    ) -> None:
        self.__taxable_event_iterator = taxable_event_iterator
        self.__acquired_lot_list = []
        self.__acquired_lot_avl: AVLTree[str, _AcquiredLotAndIndex] = AVLTree()
        self.__acquired_lot_2_partial_amount = {}

        index: int = 0
        try:
            while True:
                acquired_lot: InTransaction = next(acquired_lot_iterator)
                self.__acquired_lot_list.append(acquired_lot)
                self.__acquired_lot_avl.insert_node(
                    f"{self._get_avl_node_key(acquired_lot.timestamp, acquired_lot.internal_id)}", _AcquiredLotAndIndex(acquired_lot, index)
                )
                index += 1
        except StopIteration:
            # End of acquired_lots
            pass

        if not self.__acquired_lot_avl.root:
            raise RP2RuntimeError("Internal error: AVL tree has no root node")

    # AVL tree node keys have this format: <timestamp>_<internal_id>. The internal_id part is needed to disambiguate transactions
    # that have the same timestamp. Timestamp is in format "YYYYmmddHHMMSS.ffffff" and internal_id is padded right in a string of fixed
    # length (KEY_DISAMBIGUATOR_LENGTH).
    def _get_avl_node_key(self, timestamp: datetime, internal_id: str) -> str:
        return f"{timestamp.astimezone(timezone.utc).strftime('%Y%m%d%H%M%S.%f')}_{internal_id:0>{self.KEY_DISAMBIGUATOR_LENGTH}}"

    # This function calls _get_avl_node_key with internal_id=MAX_KEY_DISAMBIGUATOR, so that the generated key is larger than any other key
    # with the same timestamp.
    def _get_avl_node_key_with_max_disambiguator(self, timestamp: datetime) -> str:
        return self._get_avl_node_key(timestamp, self.MAX_KEY_DISAMBIGUATOR)

    @property
    def years_2_methods(self) -> AVLTree[int, AbstractAccountingMethod]:
        return self.__years_2_methods

    def _get_accounting_method(self, year: int) -> AbstractAccountingMethod:
        method = self.__years_2_methods.find_max_value_less_than(year)
        if method is None:
            raise RP2RuntimeError(f"Internal error: no accounting method assigned for year {year}")
        if not isinstance(method, AbstractAccountingMethod):
            raise RP2RuntimeError(f"Internal error: accounting method assigned for year {year} is not of type AbstractAccountingMethod: {method}")
        return method

    def _set_partial_amount(self, acquired_lot: InTransaction, amount: RP2Decimal) -> None:
        self.__acquired_lot_2_partial_amount[acquired_lot] = amount

    def get_next_taxable_event_and_amount(
        self,
        taxable_event: Optional[AbstractTransaction],
        acquired_lot: Optional[InTransaction],
        taxable_event_amount: RP2Decimal,
        acquired_lot_amount: RP2Decimal,
    ) -> TaxableEventAndAcquiredLot:
        new_acquired_lot: Optional[InTransaction] = acquired_lot
        new_acquired_lot_amount: RP2Decimal = acquired_lot_amount - taxable_event_amount if acquired_lot is not None else ZERO

        try:
            new_taxable_event: AbstractTransaction = next(self.__taxable_event_iterator)
        except StopIteration:
            raise TaxableEventsExhaustedException() from None
        new_taxable_event_amount: RP2Decimal = new_taxable_event.crypto_balance_change

        # If the new taxable event is newer than the old one (and it's not earn-typed) check if there is a newer acquired lot that
        # meets the accounting method criteria (but it's still older than the new taxable event)
        if taxable_event and taxable_event.timestamp < new_taxable_event.timestamp:
            if acquired_lot:
                self._set_partial_amount(acquired_lot, new_acquired_lot_amount)
            (_, new_acquired_lot, _, new_acquired_lot_amount) = self.get_acquired_lot_for_taxable_event(
                new_taxable_event, acquired_lot, new_taxable_event_amount, new_acquired_lot_amount
            )

        return TaxableEventAndAcquiredLot(
            taxable_event=new_taxable_event,
            acquired_lot=new_acquired_lot,
            taxable_event_amount=new_taxable_event_amount,
            acquired_lot_amount=new_acquired_lot_amount,
        )

    # After selecting the taxable event, RP2 calls this function to find the acquired_lot to pair with it. This means that the taxable
    # event can be passed to this function (which is useful for certain accounting methods)
    def get_acquired_lot_for_taxable_event(
        self,
        taxable_event: AbstractTransaction,
        acquired_lot: Optional[InTransaction],  # pylint: disable=unused-argument
        taxable_event_amount: RP2Decimal,
        acquired_lot_amount: RP2Decimal,
    ) -> TaxableEventAndAcquiredLot:
        new_taxable_event_amount: RP2Decimal = taxable_event_amount - acquired_lot_amount
        avl_result: Optional[_AcquiredLotAndIndex] = self.__acquired_lot_avl.find_max_value_less_than(
            self._get_avl_node_key_with_max_disambiguator(taxable_event.timestamp)
        )
        if avl_result is not None:
            if avl_result.acquired_lot != self.__acquired_lot_list[avl_result.index]:
                raise RP2RuntimeError("Internal error: acquired_lot incongruence in accounting logic")
            method = self._get_accounting_method(taxable_event.timestamp.year)
            lot_candidates: AcquiredLotCandidates = AcquiredLotCandidates(
                method, self.__acquired_lot_list, self.__acquired_lot_2_partial_amount, avl_result.index
            )
            acquired_lot_and_amount: Optional[AcquiredLotAndAmount] = method.seek_non_exhausted_acquired_lot(
                lot_candidates, taxable_event, new_taxable_event_amount
            )
            if acquired_lot_and_amount:
                return TaxableEventAndAcquiredLot(
                    taxable_event=taxable_event,
                    acquired_lot=acquired_lot_and_amount.acquired_lot,
                    taxable_event_amount=new_taxable_event_amount,
                    acquired_lot_amount=acquired_lot_and_amount.amount,
                )

        raise AcquiredLotsExhaustedException()
